/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "hardware/gpio.h"
#include "hardware/watchdog.h"
#include "pico/bootrom.h"
#include "pico/time.h"

#include "i2cDev.h"
#include "meter.h"

// DAC
int16_t dacValue = 0;    // D/Aコンバータの設定値

// 電圧値
double constV = 0;     // 定電圧
double dacV = 0;       // DAC電圧
double pcV = 0;        // PC電圧
double outputV = 0;    // メータ出力電圧

// TA値
double pcTA = 0;                // PCのTA値
double dacTA = 0;               // DACのTA値
double taCounter = 0;           // TAカウンター
double taList[TA_HIST_SIZE];    // TAの履歴
double taRange;                 // TAの変化量

// イベントフラグ
bool taReset = false;      // リセット
bool armMoved = false;     // トーンアーム操作
bool lcdUpdate = false;    // LCD表示更新

// 状態管理
bool   autoReset = false;    // 自動リセットモード
bool   blocking = false;     // 自動リセット抑止
bool   counting = false;     // TAカウントモード
view_t view = VIEW_1;        // LCD表示モード

/**
 * @brief 計測した電圧からTAを計算する
 */
static inline double v2TA(double v) {
    return ((v / constV) * (TA_MAX - TA_MIN)) + TA_MIN;
}

/**
 * @brief TAアクションをカウントする。
 */
void countTA() {
    static double baseTA = 0;
    if (counting) {
        if (baseTA > dacTA) {
            taCounter += (baseTA - dacTA);
        }
    }
    baseTA = dacTA;
}

/**
 * @brief イベント番号ごとの実行時間を管理し、短時間の連続実行を抑止する。
 */
bool evtEnabled(uint8_t eventNo) {
    static uint64_t nextAction[] = {0, 0, 0, 0, 0};
    bool            result = false;
    uint64_t        nowUs = time_us_64();
    if (nowUs > nextAction[eventNo]) {
        nextAction[eventNo] = nowUs + EVENT_BLOCK_uS;
        result = true;
    }
    return result;
}

/**
 * @brief 割り込みハンドラ
 */
void irqHandler(uint gpio, uint32_t events) {

    // ロータリーエンコーダー
    if (gpio == TA_RE1 || gpio == TA_RE2) {
        static const int8_t RE_DIR[] = ROTARY_DIRECTION;

        if (!taReset) {
            static uint8_t taRE1;
            static uint8_t taRE2;
            static size_t  taIndex = 0;
            size_t         lastIndex;
            uint8_t        state = 0;
            if (events == GPIO_IRQ_EDGE_RISE) {
                state = 1;
            }

            if (gpio == TA_RE1) {
                taRE1 = state;
            } else if (gpio == TA_RE2) {
                taRE2 = state;
            }

            lastIndex = taIndex;
            taIndex = ((taIndex << 2) + (taRE1 << 1) + taRE2) & 15;
            if (lastIndex != taIndex) {
                int8_t delta = RE_DIR[taIndex];
                if (delta != 0) {
                    dacValue += delta;
                    armMoved = true;
                }
            }
        }
    }

    // リセットボタン Event0
    else if (evtEnabled(0) && (gpio == BUTTON_RST) && (events == GPIO_IRQ_EDGE_FALL)) {
        // AUTO+RESET 表示切替
        if (!gpio_get(BUTTON_AUTO)) {
            view = (++view) % 3;
            lcdUpdate = true;
        }
        // リセット実行
        else {
            taReset = true;
        }
    }

    // カウントモードボタン Event1
    else if (evtEnabled(1) && (gpio == BUTTON_COUNT) && events == GPIO_IRQ_EDGE_FALL) {
        // AUTO+COUNT USB Boot
        if (!gpio_get(BUTTON_AUTO)) {
            reset_usb_boot(0, 0);
            while (1)
                ;
        }
        // RST+COUNT  TAカウンタリセット
        else if (!gpio_get(BUTTON_RST)) {
            taCounter = 0;
        }
        // カウントモード切替
        else {
            if (!counting) {
                countTA();
                counting = true;
            } else {
                counting = false;
            }
        }
        lcdUpdate = true;
    }

    // 自動モードボタン Event2
    else if (evtEnabled(2) && (gpio == BUTTON_AUTO) && (events == GPIO_IRQ_EDGE_FALL)) {
        // RESET+AUTO  マイコンリセット
        if (!gpio_get(BUTTON_RST)) {
            watchdog_enable(100, true);
            while (1)
                ;
        }
        // 自動モード切替
        else {
            autoReset = !autoReset;
            if (!autoReset) {
                blocking = false;
            }
        }
        lcdUpdate = true;
    }
}

/**
 * @brief GPIO初期化
 */
void initIO() {
    // GPIO初期化
    gpio_init(LED_PIN);
    gpio_set_dir(LED_PIN, GPIO_OUT);

    gpio_init(BUTTON_RST);
    gpio_set_dir(BUTTON_RST, GPIO_IN);
    gpio_pull_up(BUTTON_RST);
    gpio_init(BUTTON_AUTO);
    gpio_set_dir(BUTTON_AUTO, GPIO_IN);
    gpio_pull_up(BUTTON_AUTO);
    gpio_init(BUTTON_COUNT);
    gpio_set_dir(BUTTON_COUNT, GPIO_IN);
    gpio_pull_up(BUTTON_COUNT);

    gpio_init(TA_RE1);
    gpio_set_dir(TA_RE1, GPIO_IN);
    gpio_pull_up(TA_RE1);
    gpio_init(TA_RE2);
    gpio_set_dir(TA_RE2, GPIO_IN);
    gpio_pull_up(TA_RE2);

    gpio_set_irq_enabled_with_callback(BUTTON_RST, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(BUTTON_RST, GPIO_IRQ_EDGE_RISE, true, irqHandler);
    gpio_set_irq_enabled_with_callback(BUTTON_AUTO, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(BUTTON_AUTO, GPIO_IRQ_EDGE_RISE, true, irqHandler);
    gpio_set_irq_enabled_with_callback(BUTTON_COUNT, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(BUTTON_COUNT, GPIO_IRQ_EDGE_RISE, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE1, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE1, GPIO_IRQ_EDGE_RISE, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE2, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE2, GPIO_IRQ_EDGE_RISE, true, irqHandler);
}

/**
 * @brief  履歴リストをクリアする
 */
void clearList() {
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        taList[i] = pcTA;
    }
}

/**
 * @brief PCのTAを履歴を保存する。履歴からTAの変動範囲を集計する。
 */
void addTaList(double ta) {
    static uint16_t taListIndex = 0;

    taList[taListIndex] = ta;

    int    baseIndex = taListIndex + TA_HIST_SIZE;
    double maxValue = 0.0;
    int    maxIndex = 0;
    double minValue = 7.0;
    int    minIndex = 0;
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        int index = (baseIndex - i) % TA_HIST_SIZE;
        if (maxValue < taList[index]) {
            maxValue = taList[index];
            maxIndex = i;
        }
        if (minValue > taList[index]) {
            minValue = taList[index];
            minIndex = i;
        }
    }
    taRange = maxValue - minValue;
    if (minIndex < maxIndex) {
        taRange = -taRange;
    }

    ++taListIndex;
    taListIndex = taListIndex % TA_HIST_SIZE;
}

/**
 * @brief LCDを更新する。
 */
void update() {
    static char labelCount[6];
    if (counting) {
        sprintf(labelCount, "count");
    } else {
        sprintf(labelCount, "     ");
    }

    static char labelAuto[6];
    if (taReset) {
        sprintf(labelAuto, "reset");
    } else if (blocking) {
        sprintf(labelAuto, "block");
    } else if (autoReset) {
        sprintf(labelAuto, " auto");
    } else {
        sprintf(labelAuto, "     ");
    }

    static char message[30];
    lcdClear();
    if (view == VIEW_1) {    // 通常表示
        sprintf(message, "%4.2f %5.2f %s", pcTA, pcTA - dacTA, labelCount);
        lcdPrint(1, message);
        sprintf(message, "%4.2f %s%6.2f", dacTA, labelAuto, taCounter);
        lcdPrint(2, message);
    } else if (view == VIEW_2) {    // デバッグ用 TA詳細
        sprintf(message, "%6.4f %+7.4f", pcTA, pcTA - dacTA);
        lcdPrint(1, message);
        sprintf(message, "%6.4f %4d %c", dacTA, dacValue, autoReset ? 'a' : ' ');
        lcdPrint(2, message);
    } else if (view == VIEW_3) {    // デバッグ用 電圧詳細
        constV = readConstV();
        sprintf(message, "%6.4f %6.4f", pcV, constV);
        lcdPrint(1, message);
        sprintf(message, "%6.4f %6.4f %c", dacV, outputV, autoReset ? 'a' : ' ');
        lcdPrint(2, message);
    }
}

/**
 * @brief DACを0Vへ設定する
 */
void setDacZeroV() {
    gpio_put(LED_PIN, true);
    dacValue = setDacValue(DAC_MIN_VALUE);
    sleep_ms(10);

    dacV = readDacV();
    while ((dacValue < DAC_MAX_VALUE) && (dacV < 0.0)) {
        dacValue = setDacValue(dacValue + 64);
        dacV = readDacV();
    }
    while ((dacValue > DAC_MIN_VALUE) && (dacV > 0.0)) {
        dacValue = setDacValue(dacValue - 16);
        dacV = readDacV();
    }
    while ((dacValue < DAC_MAX_VALUE) && (dacV < 0.0)) {
        dacValue = setDacValue(dacValue + 4);
        dacV = readDacV();
    }
    while ((dacValue > DAC_MIN_VALUE) && (dacV > 0.0)) {
        dacValue = setDacValue(dacValue - 1);
        dacV = readDacV();
    }
    dacValue = setDacValue(dacValue + 1);

    dacV = readDacV();
    dacTA = v2TA(dacV);
    gpio_put(LED_PIN, false);
}

/**
 * @brief 針をセット位置へ移動する。
 */
void resetTA() {
    // セット位置計算の校正用
    // dacValue = (pcV / constV) * fa + fb;
    static double fa = 0.0;
    static double fb = 0.0;
    // (vRate:pcV/ConstV,value:dacValue)
    static double  p1vRate = 0.0;
    static int16_t p1value = -1.0;
    static double  p2vRate = 1.0;
    static int16_t p2value = DAC_MAX_VALUE;

    bool optimization = false;

    // 初回のみ実行
    if (p1value < 0) {
        setDacZeroV();
        p1value = dacValue;
        fa = (p1value - p2value) / (p1vRate - p2vRate);
        fb = p1value - p1vRate * (p1value - p2value) / (p1vRate - p2vRate);
        sleep_ms(20);
        pcV = readPcV();
    }

    // SET位置へジャンプ
    for (int i = 0; i < 2; i++) {
        constV = readConstV();
        dacValue = setDacValue((pcV / constV) * fa + fb);
        sleep_ms(20);
        pcV = readPcV();
    }

    // 微調整
    dacV = readDacV();
    bool    dir = pcV > dacV;
    int16_t inc = dir ? +1 : -1;
    while (taReset) {
        int16_t tValue = dacValue;
        double  tDacV = dacV;

        dacValue = setDacValue(dacValue + inc);
        dacV = readDacV();
        if (tValue == dacValue) {    // 設定可能範囲を超過
            taReset = false;
            break;
        }

        // セット位置を通り過ぎたら終了
        if (dir != (pcV > dacV)) {
            // 通り過ぎる前の方が近かったら戻す
            if (fabs(pcV - tDacV) < fabs(pcV - dacV)) {
                dacValue = setDacValue(tValue);
                dacV = readDacV();
            }
            optimization = true;
            taReset = false;
        }
    }
    dacTA = v2TA(dacV);

    // SET位置計算式を最適化
    if (optimization) {
        // 近いほうを入れ替え
        if (abs(p1value - dacValue) < abs(p2value - dacValue)) {
            // 入れ替え後の差が50以下ならば入れ替えない
            if (abs(p2value - dacValue) > 50) {
                p1vRate = pcV / constV;
                p1value = dacValue;
            }
        } else {
            // 入れ替え後の差が50以下ならば入れ替えない
            if (abs(p1value - dacValue) > 50) {
                p2vRate = pcV / constV;
                p2value = dacValue;
            }
        }
        fa = (p1value - p2value) / (p1vRate - p2vRate);
        fb = p1value - p1vRate * (p1value - p2value) / (p1vRate - p2vRate);
    }
}

/**
 * @brief 状態を確認する。
 */
void checkState() {
    if (autoReset) {
        if (blocking) {
            if (fabs(taRange) < TA_BLOCK_TH) {
                blocking = false;
            }
        } else {
            static uint64_t outRangeWait = 0;
            uint64_t        nowUS = time_us_64();
            if ((outputV <= 0.0) || ((constV * METER_MAX_RATIO) <= outputV)) {    // メータ範囲外
                if (fabs(taRange) > TA_BLOCK_TH) {                                // ボディモーション
                    blocking = true;
                    counting = false;
                    dacValue = setDacValue(DAC_MIN_VALUE);    // 不規則な動きを抑止
                } else if (nowUS > outRangeWait) {            // 待ち時間経過
                    taReset = true;                           // 自動リセット実行
                }
            } else {
                outRangeWait = nowUS + RESET_WAIT_uS;
            }
        }
    }
}

/**
 * @brief メイン
 */
int main() {
    view = VIEW_1;
    autoReset = true;
    taReset = true;

    sleep_ms(200);
    initIO();
    initI2C();

    uint64_t nextUpdateUs = 0;
    while (true) {
        while (time_us_64() < nextUpdateUs) {
            if (armMoved || taReset || lcdUpdate) {
                break;
            }
            sleep_us(500);
        }

        pcV = readPcV();
        pcTA = v2TA(pcV);
        addTaList(pcTA);
        outputV = readOutputV();
        checkState();

        if (armMoved) {
            dacValue = setDacValue(dacValue);
            dacV = readDacV();
            dacTA = v2TA(dacV);
            countTA();
            armMoved = false;
            update();
        } else if (taReset) {
            update();
            gpio_put(LED_PIN, true);
            resetTA();
            countTA();
            clearList();
            gpio_put(LED_PIN, false);
            update();
        } else if (lcdUpdate) {
            update();
            lcdUpdate = false;
        } else {
            static int32_t count = 1;
            if (count >= (UPDATE_WAIT_uS / SAMPLE_WAIT_uS)) {
                update();
                count = 1;
            } else {
                count++;
            }
            uint64_t nowUs = time_us_64();
            nextUpdateUs += SAMPLE_WAIT_uS;
            if (nextUpdateUs < nowUs) {
                nextUpdateUs = nowUs + SAMPLE_WAIT_uS;
            }
        }
    }
    return 0;
}